"""
Base module for Active Subspaces and Kernel-based Active Subspaces.

:References:

    - Paul Constantine. Active subspaces: Emerging ideas for dimension
      reduction in parameter studies, vol. 2 of SIAM Spotlights, SIAM, 2015.

    - Francesco Romor, Marco Tezzele, Andrea Lario, Gianluigi Rozza.
      Kernel-based Active Subspaces with application to CFD problems using
      Discontinuous Galerkin Method. 2020.
      arxiv: https://arxiv.org/abs/2008.12083

"""
import numpy as np
import matplotlib.pyplot as plt
from sklearn.utils.extmath import randomized_svd
from .utils import sort_eigpairs
plt.rcParams.update({'font.size': 16})


class Subspaces():
    """Active Subspaces base class

    :param str method: method to compute the AS. Possible choices are
        'exact' when the gradients are provided, or 'local' to use local linear
        models. This approach is related to the sufficient dimension reduction
        method known sometimes as the outer product of gradient method. See the
        2001 paper 'Structure adaptive approach for dimension reduction' from
        Hristache, et al.
    :param int n_boot: number of bootstrap samples. Default is 100.
    """
    def __init__(self, dim, method='exact', n_boot=100):
        self.dim = dim
        self.method = method
        self.n_boot = n_boot
        self.W1 = None
        self.W2 = None
        self.evals = None
        self.evects = None
        self.evals_br = None
        self.subs_br = None

    def _build_decompose_cov_matrix(self,
                                    gradients=None,
                                    weights=None,
                                    metric=None):
        """
        Build and decompose the covariance matrix of the gradients.

        :param numpy.ndarray gradients: n_samples-by-n_params matrix containing the
            gradient samples oriented as rows.
        :param numpy.ndarray weights: n_samples-by-1 weight vector, corresponds
            to numerical quadrature rule used to estimate matrix whose eigenspaces
            define the active subspace.
        :param numpy.ndarray metric: metric matrix output_dim-by-output-dim for
            vectorial active subspaces.

        :return: the sorted eigenvalues, and the corresponding eigenvectors.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        if self.method == 'exact' or self.method == 'local':
            if metric is not None:
                cov_matrix = np.array(
                    np.sum([
                        weights[i, 0] *
                        np.dot(gradients[i, :, :].T,
                               np.dot(metric, gradients[i, :, :]))
                        for i in range(gradients.shape[0])
                    ],
                           axis=0))
                evals, evects = sort_eigpairs(cov_matrix)
                return np.squeeze(evals), evects

            X = np.squeeze(gradients * np.sqrt(weights).reshape(-1, 1))
            n_samples, n_pars = X.shape

            # computational complexity of svd and random svd
            svd_complexity = n_samples * n_pars * self.dim
            rsvd_complexity = n_samples * n_pars * np.log(
                self.dim) + (n_samples + n_pars) * self.dim**2

            if svd_complexity > rsvd_complexity and (n_samples > 10000
                                                     or n_pars > 10000):
                singular, evects = randomized_svd(
                    M=X,
                    n_components=self.dim,
                    n_oversamples=10,
                    n_iter='auto',
                    power_iteration_normalizer='auto',
                    transpose='auto')[1:]
            else:
                singular, evects = np.linalg.svd(X, full_matrices=False)[1:]

            evals = singular**2
            return evals, evects.T

    def _compute_bootstrap_ranges(self, gradients, weights, metric=None):
        """Compute bootstrap ranges for eigenvalues and subspaces.

        An implementation of the nonparametric bootstrap that we use in
        conjunction with the subspace estimation methods to estimate the errors
        in the eigenvalues and subspaces.

        :param numpy.ndarray gradients: n_samples-by-n_params matrix containing
            the gradient samples oriented as rows.
        :param numpy.ndarray weights: n_samples-by-1 weight vector, corresponds
            to numerical quadrature rule used to estimate matrix whose
            eigenspaces define the active subspace.
        :param numpy.ndarray metric: metric matrix output_dim-by-output-dim for
            vectorial active subspaces.
        :return: array e_br is a m-by-2 matrix, first column contains
            bootstrap lower bound on eigenvalues, second column contains
            bootstrap upper bound on eigenvalues; array sub_br is a (m-1)-by-3
            matrix, first column contains bootstrap lower bound on estimated
            subspace error, second column contains estimated mean of subspace
            error (a reasonable subspace error estimate), third column contains
            estimated upper bound on subspace error.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        n_pars = gradients.shape[-1]
        n_samples = gradients.shape[0]

        svd_complexity = n_samples * n_pars * self.dim
        rsvd_complexity = n_samples * n_pars * np.log(
            self.dim) + (n_samples + n_pars) * self.dim**2

        # randomized_svd is not implemented for vectorial as yet
        if len(gradients.shape) == 2:
            if svd_complexity > rsvd_complexity and (n_samples > 10000
                                                     or n_pars > 10000):
                range_dim = self.dim
            else:
                range_dim = min(n_pars, n_samples)
        else:
            range_dim = n_pars

        e_boot = np.zeros((range_dim, self.n_boot))
        sub_dist = np.zeros((range_dim - 1, self.n_boot))

        for i in range(self.n_boot):
            gradients0, weights0 = self._bootstrap_replicate(
                gradients, weights)
            e0, W0 = self._build_decompose_cov_matrix(gradients=gradients0,
                                                      weights=weights0,
                                                      metric=metric)
            e_boot[:, i] = e0
            for j in range(range_dim - 1):
                range_diff = np.dot(self.evects[:, :j + 1].T, W0[:, j + 1:])
                sub_dist[j, i] = np.linalg.norm(range_diff, ord=2)

        # bootstrap ranges for the eigenvalues
        self.evals_br = np.hstack((np.amin(e_boot, axis=1).reshape(
            (range_dim, 1)), np.amax(e_boot, axis=1).reshape((range_dim, 1))))
        # bootstrap ranges and mean for subspace distance
        self.subs_br = np.hstack((np.amin(sub_dist, axis=1).reshape(
            (range_dim - 1, 1)), np.mean(sub_dist, axis=1).reshape(
                (range_dim - 1, 1)), np.amax(sub_dist, axis=1).reshape(
                    (range_dim - 1, 1))))

    @staticmethod
    def _bootstrap_replicate(matrix, weights):
        """
        Return a bootstrap replicate.

        A bootstrap replicate is a sampling-with-replacement strategy from a
        given data set.

        :param numpy.ndarray matrix: matrix from which will be sampled N rows. N
            corresponds to the number of rows of weights.
        :param numpy.ndarray weights: n_samples-by-1 weight vector, corresponds
            to numerical quadrature rule used to estimate matrix whose
            eigenspaces define the active subspace.
        """
        ind = np.random.randint(weights.shape[0], size=(weights.shape[0], ))

        # matrix has shape 2 if the outputs are scalar and shape 3 if they are
        # vectorial.
        if len(matrix.shape) == 2:
            return matrix[ind, :].copy(), weights[ind, :].copy()
        elif len(matrix.shape) == 3:
            return matrix[ind, :, :].copy(), weights[ind, :].copy()
        return None, None

    def fit(self, *args, **kwargs):
        """
        Abstract method to compute the active subspaces. Not implemented, it has
        to be implemented in subclasses.
        """
        raise NotImplementedError(
            'Subclass must implement abstract method {}.fit'.format(
                self.__class__.__name__))

    def transform(self, inputs):
        """
        Abstract method to map full variables to active and inactive variables.

        Points in the original input space are mapped to the active and inactive
        subspace.

        :param numpy.ndarray inputs: array n_samples-by-n_params containing the
            points in the original parameter space.
        :return: array n_samples-by-active_dim containing the mapped active variables;
            array n_samples-by-inactive_dim containing the mapped inactive
            variables.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        raise NotImplementedError('Subclass must implement abstract method '
                                  '{}.transform'.format(
                                      self.__class__.__name__))

    def inverse_transform(self, reduced_inputs, n_points):
        """
        Abstract method to find points in full space that map to reduced
        variable points. Not implemented, it has to be implemented in
        subclasses.

        :param numpy.ndarray reduced_inputs: n_samples-by-n_params matrix that
            contains points in the space of active variables.
        :param int n_points: the number of points in the original parameter
            space that are returned that map to the given active variables.
        """
        raise NotImplementedError(
            'Subclass must implement abstract method {}.inverse_transform'.
            format(self.__class__.__name__))

    def _partition(self):
        """
        Partition the eigenvectors to define the active and inactive subspaces.

        :raises: TypeError, ValueError
        """
        if not isinstance(self.dim, int):
            raise TypeError('dim should be an integer.')

        if self.dim < 1 or self.dim > self.evects.shape[1]:
            raise ValueError(
                'dim must be positive and less than the dimension of the '
                ' eigenvectors: dim = {}.'.format(self.dim))

        # allow evaluation of active eigenvectors only
        if self.evects.shape[1] < self.evects.shape[0]:
            self.W1 = self.evects[:, :self.dim]
            self.W2 = None
        elif self.evects.shape[1] == self.evects.shape[0]:
            self.W1 = self.evects[:, :self.dim]
            self.W2 = self.evects[:, self.dim:]
        else:
            raise ValueError(
                'the eigenvectors cannot have dimension less than dim = {}.'.
                format(self.dim))

    def plot_eigenvalues(self,
                         n_evals=None,
                         filename=None,
                         figsize=(8, 8),
                         title=''):
        """
        Plot the eigenvalues.

        :param int n_evals: number of eigenvalues to plot. If not provided all
            the eigenvalues will be plotted.
        :param str filename: if specified, the plot is saved at `filename`.
        :param tuple(int,int) figsize: tuple in inches defining the figure size.
            Default is (8, 8).
        :param str title: title of the plot.
        :raises: TypeError

        .. warning:: `self.fit` has to be called in advance.
        """
        if self.evals is None:
            raise TypeError('The eigenvalues have not been computed.'
                             'You have to perform the fit method.')
        if n_evals is None:
            n_evals = self.evals.shape[0]
        if n_evals > self.evals.shape[0]:
            raise TypeError('Invalid number of eigenvalues to plot.')

        plt.figure(figsize=figsize)
        plt.title(title)
        if np.amin(self.evals[:n_evals]) == 0:
            plt.semilogy(range(1, n_evals + 1),
                         self.evals[:n_evals] + np.finfo(float).eps,
                         'ko-',
                         markersize=8,
                         linewidth=2)
        else:
            plt.semilogy(range(1, n_evals + 1),
                         self.evals[:n_evals],
                         'ko-',
                         markersize=8,
                         linewidth=2)
        plt.xticks(range(1, n_evals + 1))
        plt.xlabel('Index')
        plt.ylabel('Eigenvalues')

        if self.evals_br is None:
            plt.axis([
                0, n_evals + 1, 0.1 * np.amin(self.evals[:n_evals]),
                10 * np.amax(self.evals[:n_evals])
            ])
        else:
            if np.amin(self.evals[:n_evals]) == 0:
                plt.fill_between(
                    range(1, n_evals + 1),
                    self.evals_br[:n_evals, 0] * (1 + np.finfo(float).eps),
                    self.evals_br[:n_evals, 1] * (1 + np.finfo(float).eps),
                    facecolor='0.7',
                    interpolate=True)
            else:
                plt.fill_between(range(1, n_evals + 1),
                                 self.evals_br[:n_evals, 0],
                                 self.evals_br[:n_evals, 1],
                                 facecolor='0.7',
                                 interpolate=True)
            plt.axis([
                0, n_evals + 1, 0.1 * np.amin(self.evals_br[:n_evals, 0]),
                10 * np.amax(self.evals_br[:n_evals, 1])
            ])

        plt.grid(linestyle='dotted')
        plt.tight_layout()
        
        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def plot_eigenvectors(self,
                          n_evects=None,
                          filename=None,
                          figsize=None,
                          labels=None,
                          title=''):
        """
        Plot the eigenvectors.

        :param int n_evects: number of eigenvectors to plot. Default is self.dim.
        :param str filename: if specified, the plot is saved at `filename`.
        :param tuple(int,int) figsize: tuple in inches defining the figure size.
            Default is (8, 2 * n_evects).
        :param str labels: labels for the components of the eigenvectors.
        :param str title: title of the plot.
        :raises: ValueError, TypeError

        .. warning:: `self.fit` has to be called in advance.
        """
        if self.evects is None:
            raise TypeError('The eigenvectors have not been computed.'
                             'You have to perform the fit method.')
        if n_evects is None:
            n_evects = self.dim
        if n_evects > self.evects.shape[0]:
            raise ValueError('Invalid number of eigenvectors to plot.')

        if figsize is None:
            figsize = (8, 2 * n_evects)

        n_pars = self.evects.shape[0]
        fig, axes = plt.subplots(n_evects, 1, figsize=figsize)
        fig.suptitle(title)
        # to ensure generality for subplots (1, 1)
        axes = np.array(axes)
        for i, ax in enumerate(axes.flat):
            ax.scatter(range(1, n_pars + 1),
                       self.evects[:n_pars + 1, i],
                       c='blue',
                       s=60,
                       alpha=0.9,
                       edgecolors='k')
            ax.axhline(linewidth=0.7, color='black')

            ax.set_xticks(range(1, n_pars + 1))
            if labels:
                ax.set_xticklabels(labels)

            ax.set_ylabel('Active eigenvector {}'.format(i + 1))
            ax.grid(linestyle='dotted')
            ax.axis([0, n_pars + 1, -1 - 0.1, 1 + 0.1])

        axes.flat[-1].set_xlabel('Eigenvector components')
        fig.tight_layout()
        # tight_layout does not consider suptitle so we adjust it manually
        plt.subplots_adjust(top=0.94)

        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def plot_sufficient_summary(self,
                                inputs,
                                outputs,
                                filename=None,
                                figsize=(10, 8),
                                title=''):
        """
        Plot the sufficient summary.

        :param numpy.ndarray inputs: array n_samples-by-n_params containing the
            points in the full input space.
        :param numpy.ndarray outputs: array n_samples-by-1 containing the
            corresponding function evaluations.
        :param str filename: if specified, the plot is saved at `filename`.
        :param tuple(int,int) figsize: tuple in inches defining the figure
            size. Defaults to (10, 8).
        :param str title: title of the plot.
        :raises: ValueError, TypeError

        .. warning:: `self.fit` has to be called in advance.

            Plot only available for partitions up to dimension 2.
        """
        if self.evects is None:
            raise TypeError('The eigenvectors have not been computed.'
                             'You have to perform the fit method.')

        plt.figure(figsize=figsize)
        plt.title(title)

        if self.dim == 1:
            plt.scatter(self.transform(inputs)[0],
                        outputs,
                        c='blue',
                        s=40,
                        alpha=0.9,
                        edgecolors='k')
            plt.xlabel('Active variable ' + r'$W_1^T \mathbf{\mu}}$',
                       fontsize=18)
            plt.ylabel(r'$f \, (\mathbf{\mu})$', fontsize=18)
        elif self.dim == 2:
            x = self.transform(inputs)[0]
            plt.scatter(x[:, 0],
                        x[:, 1],
                        c=outputs.reshape(-1),
                        s=60,
                        alpha=0.9,
                        edgecolors='k',
                        vmin=np.min(outputs),
                        vmax=np.max(outputs))
            plt.xlabel('First active variable', fontsize=18)
            plt.ylabel('Second active variable', fontsize=18)
            ymin = 1.1 * np.amin([np.amin(x[:, 0]), np.amin(x[:, 1])])
            ymax = 1.1 * np.amax([np.amax(x[:, 0]), np.amax(x[:, 1])])
            plt.axis('equal')
            plt.axis([ymin, ymax, ymin, ymax])
            plt.colorbar()
        else:
            raise ValueError(
                'Sufficient summary plots cannot be made in more than 2 '
                'dimensions.')

        plt.grid(linestyle='dotted')
        plt.tight_layout()

        if filename:
            plt.savefig(filename)
        else:
            plt.show()



"""
Module for the feature map class

:References:

    - Francesco Romor, Marco Tezzele, Andrea Lario, Gianluigi Rozza.
      Kernel-based Active Subspaces with application to CFD problems using
      Discontinuous Galerkin Method. 2020.
      arxiv: https://arxiv.org/abs/2008.12083

"""
from functools import partial
import numpy as np
from scipy.optimize import brute, dual_annealing
import GPyOpt
from .projection_factory import ProjectionFactory


class FeatureMap():
    """Feature map class.

    :param str distr: name of the spectral distribution to sample from the
        matrix.
    :param numpy.ndarray bias: n_features dimensional bias. It is sampled from a
        Unifrom distribution on the interval [0, 2*PI].
    :param int input_dim: dimension of the input space.
    :param int n_features: dimension of the Reproducing Kernel Hilbert Space.
    :param list params: number of hyperparameters of the spectral distribution.
    :param int sigma_f: multiplicative constant of the feature
        map. Usually it is set as the empirical variance of the outputs.

    :cvar callable fmap: feature map used to project the input samples to the RKHS.
        Default value is rff_map.
    :cvar callable fjac: Jacobian matrix of fmap. Default value is rff_jac, the
        analytical Jacobian of fmap.
    :cvar numpy.ndarray _pr_matrix: n_features-by-input_dim projection
        matrix, whose rows are sampled from the spectral distribution distr.

    :raises TypeError
    """
    def __init__(self, distr, bias, input_dim, n_features, params, sigma_f):
        if callable(distr):
            self.distr = distr
        elif isinstance(distr, str):
            self.distr = ProjectionFactory(distr)
        else:
            raise TypeError('`distr` is not valid.')

        self.bias = bias
        self.input_dim = input_dim
        self.n_features = n_features
        self.params = params
        self.sigma_f = sigma_f
        self.fmap = rff_map
        self.fmap_jac = rff_jac

        self._pr_matrix = None

    @property
    def pr_matrix(self):
        """
        Get the projection matrix.

        :return: the projection matrix.
        :rtype: numpy.ndarray
        """
        return self._pr_matrix

    def _compute_pr_matrix(self):
        """
        Sample the projection matrixx from the spectral distribution.

        :return: the projection matrix.
        :rtype: numpy.ndarray
        """
        return self.distr(self.input_dim, self.n_features, self.params)

    def compute_fmap(self, inputs):
        """
        Evaluate the feature map at inputs.

        :param numpy.ndarray inputs: the inputs to project on the RKHS.
        :return: the n_features dimensional projection of the inputs.
        :rtype: numpy.ndarray
        """
        if self._pr_matrix is None:
            self._pr_matrix = self._compute_pr_matrix()
        return self.fmap(inputs, self._pr_matrix, self.bias, self.n_features,
                         self.sigma_f)

    def compute_fmap_jac(self, inputs):
        """
        Evaluate the Jacobian matrix of feature map at inputs.

        :param numpy.ndarray inputs: the inputs at which compute the Jacobian
            matrix of the feature map.
        :return: the n_features-by-input_dim dimensional Jacobian of the
            feature map at the inputs.
        :rtype: numpy.ndarray
        """
        if self._pr_matrix is None:
            self._pr_matrix = self._compute_pr_matrix()
        return self.fmap_jac(inputs, self._pr_matrix, self.bias,
                             self.n_features, self.sigma_f)

    def tune_pr_matrix(self,
                       func,
                       bounds,
                       method=None,
                       maxiter=50,
                       save_file=False,
                       fn_args={}):
        """
        Tune the parameters of the spectral distribution. Three methods are
        available: log-grid-search (brute), annealing (dual_annealing) and
        Bayesian stochastic optimization (bso) from GpyOpt. The default object
        function to optimize is athena.utils.average_rrmse, which uses a
        cross-validation procedure from athena.utils, see Example and tutorial 06_kernel-based_AS.

        :Example:
        >>> from athena.kas import KernelActiveSubspaces
        >>> from athena.feature_map import FeatureMap
        >>> from athena.projection_factory import ProjectionFactory
        >>> from athena.utils import CrossValidation, average_rrmse
        >>> input_dim, output_dim, n_samples = 2, 1, 30
        >>> n_features, n_params = 10, 1
        >>> xx = np.ones((n_samples, input_dim))
        >>> f = np.ones((n_samples, output_dim))
        >>> df = np.ones((n_samples, output_dim, input_dim))
        >>> fm = FeatureMap(distr='laplace',
                            bias=np.random.uniform(0, 2 * np.pi, n_features),
                            input_dim=input_dim,
                            n_features=n_features,
                            params=np.zeros(n_params),
                            sigma_f=f.var())
        >>> kss = KernelActiveSubspaces(feature_map=fm, dim=1, n_features=n_features)
        >>> csv = CrossValidation(inputs=xx,
                                outputs=f.reshape(-1, 1),
                                gradients=df.reshape(n_samples, output_dim, input_dim),
                                folds=3,
                                subspace=kss)
        >>> best = fm.tune_pr_matrix(func=average_rrmse,
                                    bounds=[slice(-2, 1, 0.2) for i in range(n_params)],
                                    args=(csv, ),
                                    method='bso',
                                    maxiter=20,
                                    save_file=False)

        :param callable func: the objective function to be minimized.
            Must be in the form f(x, *args), where x is the argument in the
            form of a 1-D array and args is a tuple of any additional fixed
            parameters needed to completely specify the function.
        :param tuple bounds: each component of the bounds tuple must be a
            slice tuple of the form (low, high, step). It defines bounds for
            the objective function parameter in a logarithmic scale. Step will
            be ignored for 'dual_annealing' method.
        :param tuple args: any additional fixed parameters needed to
            completely specify the objective function.
        :param str method: method used to optimize the objective function.
            Possible values are 'brute', or 'dual_annealing'.
            Default value is None, and the choice is made automatically wrt
            the dimension of `self.params`. If the dimension is less than 4
            brute force is used, otherwise a traditional Generalized
            Simulated Annealing will be performed with no local search
            strategy applied.
        :param int maxiter: the maximum number of global search iterations.
            Default value is 50.
        :param bool save_file: True to save the optimal projection matrix
        :param dict fn_args: dictionary of arguments passed to func.
        :raises: ValueError
        :return: list that records the best score and the best projection
            matrix. The initial values are 0.8 and a n_features-by-input_dim
            numpy.ndarray of zeros.
        :rtype: list
        """
        best = [0.8, np.zeros((self.n_features, self.input_dim))]

        if method is None:
            if len(self.params) < 4:
                method = 'brute'
            else:
                method = 'dual_annealing'

        if method == 'brute':
            self.params = brute(func=func,
                                ranges=bounds,
                                args=(
                                    best,
                                    *tuple(fn_args.values()),
                                ),
                                finish=None)
        elif method == 'dual_annealing':
            bounds_list = [[bound.start, bound.stop] for bound in bounds]
            self.params = 10**dual_annealing(func=func,
                                             bounds=bounds_list,
                                             args=(
                                                 best,
                                                 *tuple(fn_args.values()),
                                             ),
                                             maxiter=maxiter,
                                             no_local_search=False).x
        elif method == 'bso':
            bounds = [{
                'name': 'var_' + str(i),
                'type': 'continuous',
                'domain': [bound.start, bound.stop]
            } for i, bound in enumerate(bounds)]
            func_obj = partial(func, best=best, **fn_args)
            bopt = GPyOpt.methods.BayesianOptimization(func_obj,
                                                       domain=bounds,
                                                       model_type='GP',
                                                       acquisition_type='EI',
                                                       exact_feval=True)
            bopt.run_optimization(max_iter=maxiter,
                                  max_time=3600,
                                  eps=1e-16,
                                  verbosity=False)
            self.params = 10**bopt.x_opt
        else:
            raise ValueError(
                "Method argument can only be 'brute' or 'dual_annealing' or 'bso'."
            )

        self._pr_matrix = best[1]
        self.params = best[0]

        if save_file:
            np.save("opt_pr_matrix", best[1])

        return best


def rff_map(inputs, pr_matrix, bias, n_features, sigma_f):
    """
    Random Fourier Features map. It can be vectorized for inputs of shape n_samples-by-input_dim.

    :param numpy.ndarray inputs: input_dim dimensional inputs to project to the RKHS.
    :param numpy.ndarray pr_matrix: n_features-by-input_dim projection matrix,
        whose rows are sampled from the spectral distribution.
    :param numpy.ndarray bias: n_features dimensional bias. It is sampled from a
        Unifrom distribution on the interval [0, 2*PI].
    :param int n_features: dimension of the RKHS
    :param int sigma_f: multiplicative term representing the empirical variance
        the outptus.
    :return: n_features dimensional projection of the inputs to the RKHS
    :rtype: numpy.ndarray
    """
    return np.sqrt(2 / n_features) * sigma_f * np.cos(
        np.dot(inputs, pr_matrix.T) + bias.reshape(1, -1))


def rff_jac(inputs, pr_matrix, bias, n_features, sigma_f):
    """
    Random Fourier Features map's Jacobian. It can be vectorized for inputs of shape n_samples-by-input_dim.

    :param numpy.ndarray inputs: input_dim dimensional inputs to project to the RKHS.
    :param numpy.ndarray pr_matrix: n_features-by-input_dim projection matrix,
        whose rows are sampled from the spectral distribution.
    :param numpy.ndarray bias: n_features dimensional bias. It is sampled from a
        Unifrom distribution on the interval [0, 2*PI].
    :param int n_features: dimension of the RKHS
    :param int sigma_f: multiplicative term representing the empirical variance
        the outptus.
    :return: n_features-by-input_dim dimensional projection of the inputs to the RKHS
    :rtype: numpy.ndarray
    """
    return (np.sqrt(2 / n_features) * sigma_f *
            (-1) * np.sin(np.dot(inputs, pr_matrix.T) + bias)).reshape(
                inputs.shape[0], n_features, 1) * pr_matrix



"""
Active Subspaces module.

:References:

    - Paul Constantine. Active subspaces: Emerging ideas for dimension
      reduction in parameter studies, vol. 2 of SIAM Spotlights, SIAM, 2015.

    - Constantine et al. Python Active-subspaces Utility Library,
      Journal of Open Source Software, 1(5), 79, 2016.

"""
import numpy as np
from .subspaces import Subspaces
from scipy.linalg import null_space
import types

from .utils import (Normalizer, initialize_weights, linear_program_ineq,
                    local_linear_gradients)


class ActiveSubspaces(Subspaces):
    """Active Subspaces class

    :param int dim: dimension of the active subspace.
    :param str method: method to compute the AS. Possible choices are
        'exact' when the gradients are provided, or 'local' to use local linear
        models. This approach is related to the sufficient dimension reduction
        method known sometimes as the outer product of gradient method. See the
        2001 paper 'Structure adaptive approach for dimension reduction' from
        Hristache, et al.
    :param int n_boot: number of bootstrap samples. Default is 100.
    """
    def __init__(self, dim, method='exact', n_boot=100):
        super().__init__(dim, method, n_boot)

    @property
    def activity_scores(self):
        """
        Return the activity scores as defined in Constantine and Diaz, Global
        sensitivity metrics from active subspaces, arxiv.org/abs/1510.04361
        Equation (21).
        
        :return: array with the activity score of each parameter.
        :rtype: numpy.ndarray
        :raises: TypeError

        .. warning:: `self.fit` has to be called in advance.
        """
        if self.W1 is None:
            raise TypeError('The eigenvectors have not been computed and '
                            'partitioned. You have to perform the fit method.')

        return np.power(self.W1, 2).dot(self.evals[:self.dim])

    def fit(self,
            inputs=None,
            outputs=None,
            gradients=None,
            weights=None,
            metric=None):
        """
        Compute the active subspaces given the gradients of the model function
        wrt the input parameters, or given the input/outputs couples. Only two
        methods are available: 'exact' and 'local'.

        :param numpy.ndarray inputs: input parameters oriented as rows.
        :param numpy.ndarray outputs: corresponding outputs oriented as rows.
        :param numpy.ndarray gradients: n_samples-by-n_params matrix containing
            the gradient samples oriented as rows. If frequent directions needed
            to be performed, gradients is an object of GeneratorType.
        :param numpy.ndarray weights: n_samples-by-1 weight vector, corresponds
            to numerical quadrature rule used to estimate matrix whose
            eigenspaces define the active subspace.
        :param numpy.ndarray metric: metric matrix output_dim-by-output-dim for
            vectorial active subspaces.
        :raises: TypeError

        :Example:

            >>> # inputs shape is n_samples-by-n_params
            >>> # outputs shape is n_samples-by-1
            >>> # gradients shape is n_samples-by-n_params

            >>> # if gradients are available use the 'exact' method:
            >>> ss1 =  ActiveSubspaces(dim=1, method='exact', n_boot=150)
            >>> ss1.fit(gradients=gradients)

            >>> # for the frequent direction method to compute the eigenpairs
            >>> # you need to pass a generator to gradients:
            >>> gradients_gen = (grad for grad in gradients)
            >>> ss2 =  ActiveSubspaces(dim=2, method='exact')
            >>> ss2.fit(gradients=gradients_gen)

            >>> # if no gradients are available use the 'local' method:
            >>> ss3 =  ActiveSubspaces(dim=1, method='local', n_boot=150)
            >>> ss3.fit(inputs=inputs, outputs=outputs)
        """
        if self.method == 'exact':
            if gradients is None:
                raise TypeError('gradients argument is None.')

        # estimate active subspace with local linear models.
        if self.method == 'local':
            if inputs is None or outputs is None:
                raise TypeError('inputs or outputs argument is None.')
            gradients = local_linear_gradients(inputs=inputs,
                                               outputs=outputs,
                                               weights=weights)[0]

        if isinstance(gradients, types.GeneratorType):
            self.evals, self.evects = self._frequent_directions(gradients)
        else:
            if weights is None or self.method == 'local':
                # use the new gradients to compute the weights, otherwise
                # dimension mismatch accours.
                weights = initialize_weights(gradients)

            if len(gradients.shape) == 3 and metric is None:
                metric = np.diag(np.ones(gradients.shape[1]))

            self.evals, self.evects = self._build_decompose_cov_matrix(
                gradients=gradients, weights=weights, metric=metric)

            self._compute_bootstrap_ranges(gradients, weights, metric=metric)
        self._partition()

    def transform(self, inputs):
        """
        Map full variables to active and inactive variables.

        Points in the original input space are mapped to the active and inactive
        subspace.

        :param numpy.ndarray inputs: array n_samples-by-n_params containing the
            points in the original parameter space.
        :return: array n_samples-by-active_dim containing the mapped active
            variables;
            array n_samples-by-inactive_dim containing the mapped inactive
            variables.
        :rtype: numpy.ndarray, numpy.ndarray
        :raises: TypeError
        """
        if self.W1 is None:
            raise TypeError('the active subspace has not been evaluated.')

        # allow evaluation of active variables only
        if self.W2 is None:
            active = np.dot(inputs, self.W1)
            inactive = None
        else:
            active = np.dot(inputs, self.W1)
            inactive = np.dot(inputs, self.W2)

        return active, inactive

    def inverse_transform(self, reduced_inputs, n_points=1):
        """
        Map the points in the active variable space to the original parameter
        space.

        :param numpy.ndarray reduced_inputs: n_samples-by-dim matrix that
            contains points in the space of active variables.
        :param int n_points: the number of points in the original parameter
            space that are returned that map to the given active variables.
            Defaults to 1.
        :return: (n_samples * n_points)-by-n_params matrix that contains
            points in the original parameter space, (n_samples *
            n_points)-by-n_params matrix that contains integer indices. These
            indices identify which rows of the previous matrix (the full
            parameters) map to which rows of the active variables matrix.
        :rtype: numpy.ndarray, numpy.ndarray
        :raises: TypeError

        .. note:: The inverse map depends critically on the
            `self._sample_inactive` method.
        """
        if self.W1 is None:
            raise TypeError('the active subspace has not been evaluated.')

        # the inactive eigenvectors are needed
        if self.W2 is None:
            self.W2 = null_space(self.W1).T

        inactive_swap = np.array([
            self._sample_inactive(red_inp, n_points)
            for red_inp in reduced_inputs
        ])
        inactive_inputs = np.swapaxes(inactive_swap, 1, 2)

        inputs, indices = self._rotate_x(reduced_inputs, inactive_inputs)
        return inputs, indices

    def _sample_inactive(self, reduced_input, n_points):
        """
        Sample inactive variables.

        Sample values of the inactive variables for a fixed value of the active
        variables when the original variables are bounded by a hypercube.

        :param numpy.ndarray reduced_input: the value of the active variables.
        :param int n_points: the number of inactive variable samples,
        :return: n_points-by-(inactive_dim) matrix that contains values of the
            inactive variable that correspond to the given `reduced_input`.
        :rtype: numpy.ndarray

        .. note:: The trick here is to sample the inactive variables z so that
            -1 <= W1*y + W2*z <= 1, where y is the given value of the active
            variables. In other words, we need to sample z such that it respects
            the linear inequalities W2*z <= 1 - W1*y, -W2*z <= 1 + W1*y. These
            inequalities define a polytope in R^(inactive_dim). We want to
            sample N points uniformly from the polytope. This function first
            tries a simple rejection sampling scheme, which finds a bounding
            hyperbox for the polytope, draws points uniformly from the bounding
            hyperbox, and rejects points outside the polytope. If that method
            does not return enough samples, the method tries a "hit and run"
            method for sampling from the polytope. If that does not work, it
            returns an array with `N` copies of a feasible point computed as the
            Chebyshev center of the polytope.
        """
        Z = self._rejection_sampling_inactive(reduced_input, n_points)
        if Z is None:
            Z = self._hit_and_run_inactive(reduced_input, n_points)
        return Z

    def _compute_A_b(self, reduced_input):
        """
        Compute the matrix A and the vector b to build a box around the
        inactive subspace for uniform sampling.

        :param numpy.ndarray reduced_input: the value of the active variables.
        :return: matrix A, and vector b.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        s = np.dot(self.W1, reduced_input).reshape((-1, 1))
        A = np.vstack((self.W2, -1 * self.W2))
        b = np.vstack((-1 - s, -1 + s)).reshape((-1, 1))
        return A, b

    def _rejection_sampling_inactive(self, reduced_input, n_points):
        """
        A rejection sampling method for sampling the from a polytope.

        :param numpy.ndarray reduced_input: the value of the active variables.
        :param int n_points: the number of inactive variable samples.
        :return: n_points-by-inactive_dim matrix that contains values of the
            inactive variable that correspond to the given `reduced_input`.
        :rtype: numpy.ndarray
        """
        m, n = self.W1.shape
        inactive_dim = m - n
        # Build a box around z for uniform sampling
        A, b = self._compute_A_b(reduced_input)
        lbox, ubox = np.zeros((1, inactive_dim)), np.zeros((1, inactive_dim))
        for i in range(inactive_dim):
            clb = np.zeros((inactive_dim, 1))
            clb[i, 0] = 1.0
            lbox[0, i] = linear_program_ineq(clb, A, b)[i, 0]
            cub = np.zeros((inactive_dim, 1))
            cub[i, 0] = -1.0
            ubox[0, i] = linear_program_ineq(cub, A, b)[i, 0]
        bn = Normalizer(lbox, ubox)
        Zbox = bn.inverse_transform(
            np.random.uniform(-1.0, 1.0, size=(50 * n_points, inactive_dim)))
        ind = np.all(np.dot(A, Zbox.T) >= b, axis=0)

        if np.sum(ind) >= n_points:
            Z = Zbox[ind, :]
            return Z[:n_points, :].reshape(n_points, inactive_dim)
        return None

    def _hit_and_run_inactive(self, reduced_input, n_points):
        """
        A hit and run method for sampling the inactive variables from a
        polytope.

        :param numpy.ndarray reduced_input: the value of the active variables.
        :param int n_points: the number of inactive variable samples.
        :return: n_points-by-(inactive_dim) matrix that contains values of the
            inactive variable that correspond to the given `reduced_input`.
        :rtype: numpy.ndarray
        """
        m, n = self.W1.shape
        inactive_dim = m - n

        # Get an initial feasible point using the Chebyshev center. Huge props
        # to David Gleich for the Chebyshev center.
        s = np.dot(self.W1, reduced_input).reshape((m, 1))
        normW2 = np.sqrt(np.sum(np.power(self.W2, 2), axis=1)).reshape((m, 1))
        A = np.hstack((np.vstack(
            (self.W2, -self.W2.copy())), np.vstack((normW2, normW2.copy()))))
        b = np.vstack((1 - s, 1 + s)).reshape((2 * m, 1))
        c = np.zeros((inactive_dim + 1, 1))
        c[-1] = -1.0

        zc = linear_program_ineq(c, -1 * A, -b)
        z0 = zc[:-1].reshape((inactive_dim, 1))

        # define the polytope A >= b
        A, b = self._compute_A_b(reduced_input)

        # tolerance
        eps0 = 1e-6 / 4.0

        Z = np.zeros((n_points, inactive_dim))
        for i in range(n_points):
            # random direction
            bad_dir = True
            count, maxcount = 0, 50
            while bad_dir:
                d = np.random.normal(size=(inactive_dim, 1))
                bad_dir = np.any(np.dot(A, z0 + eps0 * d) <= b)
                count += 1
                if count >= maxcount:
                    Z[i:, :] = np.tile(z0, (1, n_points - i)).T
                    return Z

            # find constraints that impose lower and upper bounds on eps
            f, g = b - np.dot(A, z0), np.dot(A, d)

            # find an upper bound on the step
            min_ind = np.logical_and(g <= 0,
                                     f < -np.sqrt(np.finfo(np.float).eps))
            eps_max = np.amin(f[min_ind] / g[min_ind])

            # find a lower bound on the step
            max_ind = np.logical_and(g > 0,
                                     f < -np.sqrt(np.finfo(np.float).eps))
            eps_min = np.amax(f[max_ind] / g[max_ind])

            # randomly sample eps
            eps1 = np.random.uniform(eps_min, eps_max)

            # take a step along d
            z1 = z0 + eps1 * d
            Z[i, :] = z1.reshape(-1, )

            # update temp var
            z0 = z1.copy()

        return Z

    def _rotate_x(self, reduced_inputs, inactive_inputs):
        """
        A convenience function for rotating subspace coordinates to x space.

        :param numpy.ndarray reduced_input: the value of the active variables.
        :param numpy.ndarray inactive_inputs: the value of the inactive
            variables.
        :return: (n_samples * n_points)-by-n_params matrix that contains points
            in the original parameter space, (n_samples * n_points)-by-n_params
            matrix that contains integer indices. These indices identify which
            rows of the previous matrix (the full parameters) map to which rows
            of the active variables matrix.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        NY, n = reduced_inputs.shape
        N = inactive_inputs.shape[2]
        m = n + inactive_inputs.shape[1]

        YY = np.tile(reduced_inputs.reshape((NY, n, 1)), (1, 1, N))
        YZ = np.concatenate((YY, inactive_inputs), axis=1).transpose(
            (1, 0, 2)).reshape((m, N * NY)).transpose((1, 0))
        inputs = np.dot(YZ, self.evects.T).reshape((N * NY, m))
        indices = np.kron(np.arange(NY), np.ones(N)).reshape((N * NY, 1))
        return inputs, indices

    def _frequent_directions(self, gradients):
        """
        Function that performs the frequent directions algorithm for 
        matrix sketching. For more details about the method, see
        "Frequent directions: Simple and deterministic matrix
        sketching." Ghashami, Mina, et al.
        SIAM Journal on Computing 45.5 (2016): 1762-1792.
        doi: https://doi.org/10.1137/15M1009718

        :param iterable gradients: generator for spatial gradients.
        :return the sorted eigenvalues and the corresponding eigenvectors for
            the reduced matrix.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        s = np.array([next(gradients) for i in range(self.dim)]).T
        for grad in gradients:
            evects, sigma = np.linalg.svd(s, full_matrices=False)[:2]
            s = np.dot(
                evects,
                np.sqrt(np.diag(sigma**2) - (sigma[-1]**2) * np.eye(self.dim)))
            s[:, -1] = grad
        evals = sigma**2
        return evals, evects



"""
Module for Nonlinear Level-set Learning (NLL).

:References:

    - Guannan Zhang, Jiaxin Zhang, Jacob Hinkle.
      Learning nonlinear level sets for dimensionality reduction in function
      approximation. NeurIPS 2019, 13199-13208.
      arxiv: https://arxiv.org/abs/1902.10652

"""
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt

# Define the overall data type
torch.set_default_tensor_type(torch.DoubleTensor)


class NonlinearLevelSet():
    """Nonlinear Level Set class. It is implemented as a Reversible neural
    networks (RevNet).

    :param int n_layers: number of layers of the RevNet.
    :param int active_dim: number of active dimensions.
    :param float lr: learning rate.
    :param int epochs: number of ephocs.
    :param float dh: so-called time step of the RevNet. Default is 0.25.
    :param `torch.optim.Optimizer` optimizer: optimizer used in the training of
        the RevNet. Its argument are passed in the dict optim_args when
        :py:meth:`train` is called.
    :param `torch.optim.lr_scheduler._LRScheduler` scheduler: scheduler used in
        the training of the RevNet. Its argument are passed in the dict
        scheduler_args when :py:meth:`train` is called. Default is None.

    :cvar `BackwardNet` backward: backward net of the RevNet. See
        :class:`BackwardNet` class in :py:mod:`nll` module.
    :cvar `ForwardNet` forward: forward net of the RevNet. See
        :class:`ForwardNet` class in :py:mod:`nll` module.
    :cvar list loss_vec: list containg the loss at every epoch.
    """
    def __init__(self,
                 n_layers,
                 active_dim,
                 lr,
                 epochs,
                 dh=0.25,
                 optimizer=optim.Adam,
                 scheduler=None):
        self.n_layers = n_layers
        self.active_dim = active_dim
        self.lr = lr
        self.epochs = epochs
        self.dh = dh
        if issubclass(optimizer, optim.Optimizer):
            self.optimizer = optimizer

        if scheduler and issubclass(scheduler,
                                    torch.optim.lr_scheduler._LRScheduler):
            self.scheduler = scheduler
        else:
            self.scheduler = None

        self.backward = None
        self.forward = None
        self.loss_vec = []

    def train(self,
              inputs,
              gradients,
              outputs=None,
              interactive=False,
              target_loss=0.0001,
              optim_args={},
              scheduler_args={}):
        """
        Train the whole RevNet.

        :param torch.Tensor inputs: DoubleTensor n_samples-by-n_params
            containing the points in the full input space.
        :param torch.Tensor gradients: DoubleTensor n_samples-by-n_params
            containing the gradient samples wrt the input parameters.
        :param numpy.ndarray outputs: array n_samples-by-1 containing the
            corresponding function evaluations. Needed only for the interactive
            mode. Default is None.
        :param bool interactive: if True a plot with the loss function decay,
            and the sufficient summary plot will be showed and updated every
            10 epochs, and at the last epoch. Default is False.
        :param float target_loss: loss threshold. Default is 0.0001.
        :param dict optim_args: dictionary passed to the optimizer.
        :param dict scheduler_args: dictionary passed to the scheduler.
        :raises: ValueError: in interactive mode outputs must be provided for
            the sufficient summary plot.
        """
        if interactive:
            if outputs is None:
                raise ValueError(
                    'outputs in interactive mode have to be provided!')
            fig = plt.figure(figsize=(12, 5))
            ax1 = fig.add_subplot(1, 2, 1)
            ax2 = fig.add_subplot(1, 2, 2)
            ax2.set_title('Loss function decay')
            ax2.grid(linestyle='dotted')
            plt.ion()
            plt.show()

        # Build the forward network
        self.forward = ForwardNet(n_params=inputs.shape[1],
                                  n_layers=self.n_layers,
                                  dh=self.dh,
                                  active_dim=self.active_dim)
        # Initialize the gradient
        self.forward.zero_grad()
        optimizer = self.optimizer(self.forward.parameters(), self.lr,
                                   **optim_args)

        # Initialize scheduler if present
        if self.scheduler:
            sched = self.scheduler(optimizer, **scheduler_args)

        # Training
        for i in range(self.epochs):
            optimizer.zero_grad()
            mapped_inputs = self.forward(inputs)
            loss = self.forward.customized_loss(inputs, mapped_inputs,
                                                gradients)

            if i % 10 == 0 or i == self.epochs - 1:
                print('epoch = {}, loss = {}'.format(i, loss))
                if interactive:
                    ax1.cla()
                    ax1.set_title('Sufficient summary plot')
                    ax1.plot(mapped_inputs.detach().numpy()[:, 0], outputs,
                             'bo')
                    ax1.grid(linestyle='dotted')
                    ax2.plot(i, loss.detach().numpy(), 'ro')
                    plt.draw()
                    plt.pause(0.00001)

                self.loss_vec.append(loss.detach().numpy())
                # Build the inverse network based on the trained forward network
                self.backward = BackwardNet(n_params=inputs.shape[1],
                                            n_layers=self.n_layers,
                                            dh=self.dh)
                self.backward.zero_grad()

                for j in range(self.backward.n_layers):
                    name_y = 'fc{}_y'.format(j + 1)
                    name_z = 'fc{}_z'.format(j + 1)
                    getattr(self.backward, name_y).weight = torch.nn.Parameter(
                        getattr(self.forward, name_y).weight)
                    getattr(self.backward, name_z).weight = torch.nn.Parameter(
                        getattr(self.forward, name_z).weight)
                    getattr(self.backward, name_y).bias = torch.nn.Parameter(
                        getattr(self.forward, name_y).bias)
                    getattr(self.backward, name_z).bias = torch.nn.Parameter(
                        getattr(self.forward, name_z).bias)

                if torch.mean(
                        torch.abs(
                            torch.add(-1 * inputs,
                                      self.backward(
                                          self.forward(inputs))))) > 1e-5:
                    print('self.backward is wrong!')
                    print(
                        torch.mean(
                            torch.abs(
                                torch.add(-1 * inputs,
                                          self.backward(
                                              self.forward(inputs))))))

                if loss < target_loss:
                    break

            loss.backward()
            optimizer.step()

            if self.scheduler:
                sched.step()

        if interactive:
            plt.ioff()
            plt.show()

    def plot_sufficient_summary(self,
                                inputs,
                                outputs,
                                filename=None,
                                figsize=(10, 8),
                                title=''):
        """
        Plot the sufficient summary.

        :param torch.Tensor inputs: DoubleTensor n_samples-by-n_params
            containing the points in the full input space.
        :param numpy.ndarray outputs: array n_samples-by-1 containing the
            corresponding function evaluations.
        :param str filename: if specified, the plot is saved at `filename`.
        :param tuple(int,int) figsize: tuple in inches defining the figure
            size. Defaults to (10, 8).
        :param str title: title of the plot.
        :raises: ValueError

        .. warning::
            Plot only available for active dimensions up to 1.
        """
        plt.figure(figsize=figsize)
        plt.title(title)

        if self.active_dim == 1:
            reduced_inputs = self.forward(inputs)[:, 0]
            plt.plot(reduced_inputs.detach().numpy(), outputs, 'bo')
            plt.xlabel('Reduced input')
            plt.xlabel('Output')
        else:
            raise ValueError(
                'Sufficient summary plots cannot be made in more than 1 ' \
                'dimension.'
            )

        plt.grid(linestyle='dotted')

        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def plot_loss(self, filename=None, figsize=(10, 8), title=''):
        """
        Plot the loss function decay.

        :param str filename: if specified, the plot is saved at `filename`.
        :param tuple(int,int) figsize: tuple in inches defining the figure
            size. Defaults to (10, 8).
        :param str title: title of the plot.
        """
        plt.figure(figsize=figsize)
        plt.title(title)
        x_range = list(range(1, self.epochs + 1, 10)) + [self.epochs]
        plt.plot(x_range, self.loss_vec, 'b-')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.grid(linestyle='dotted')
        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def save_forward(self, outfile):
        """
        Save the forward map for future inference.

        :param str outfile: filename of the net to save.
            Use either .pt or .pth. See notes below.

        .. note::
            A common PyTorch convention is to save models using either a .pt or
            .pth file extension.
        """
        torch.save(self.forward.state_dict(), outfile)

    def load_forward(self, infile, n_params):
        """
        Load the forward map for inference.

        :param str infile: filename of the saved net to load. See notes below.
        :param int n_params: number of input parameters.

        .. note::
            A common PyTorch convention is to save models using either a .pt or
            .pth file extension.
        """
        self.forward = ForwardNet(n_params, self.n_layers, self.dh,
                                  self.active_dim)
        self.forward.load_state_dict(torch.load(infile))
        self.forward.eval()

    def save_backward(self, outfile):
        """
        Save the backward map for future inference.

        :param str outfile: filename of the net to save.
            Use either .pt or .pth. See notes below.

        .. note::
            A common PyTorch convention is to save models using either a .pt or
            .pth file extension.
        """
        torch.save(self.backward.state_dict(), outfile)

    def load_backward(self, infile, n_params):
        """
        Load the backward map for inference.

        :param str infile: filename of the saved net to load. See notes below.
        :param int n_params: number of input parameters.

        .. note::
            A common PyTorch convention is to save models using either a .pt or
            .pth file extension.
        """
        self.backward = BackwardNet(n_params, self.n_layers, self.dh)
        self.backward.load_state_dict(torch.load(infile))
        self.backward.eval()


class ForwardNet(nn.Module):
    """
    Forward net class. It is part of the RevNet.

    :param int n_params: number of input parameters.
    :param int n_layers: number of layers of the RevNet.
    :param float dh: so-called time step of the RevNet.
    :param int active_dim: number of active dimensions.

    :cvar slice omega: a slice object indicating the active dimension to keep.
        For example to keep the first two dimension `omega = slice(2)`. It is
        automatically set with `active_dim`.
    """
    def __init__(self, n_params, n_layers, dh, active_dim):
        super().__init__()
        self.n_params = n_params // 2
        self.n_layers = n_layers
        self.dh = dh
        self.omega = slice(active_dim)

        for i in range(self.n_layers):
            setattr(self, 'fc{}_y'.format(i + 1),
                    nn.Linear(self.n_params, 2 * self.n_params))
            setattr(self, 'fc{}_z'.format(i + 1),
                    nn.Linear(self.n_params, 2 * self.n_params))

    def forward(self, inputs):
        """
        Maps original inputs to transformed inputs.

        :param torch.Tensor inputs: DoubleTensor n_samples-by-n_params
            containing the points in the original full input space.
        :return mapped_inputs: DoubleTensor n_samples-by-n_params with
            the nonlinear transformed inputs.
        :rtype: torch.Tensor
        """
        bb = torch.split(inputs, self.n_params, dim=1)
        vars()['var0_y'] = torch.clone(bb[0])
        vars()['var0_z'] = torch.clone(bb[1])

        for i in range(self.n_layers):
            name_y = 'fc{}_y'.format(i + 1)
            name_z = 'fc{}_z'.format(i + 1)
            var_y0 = 'var{}_y'.format(i)
            var_z0 = 'var{}_z'.format(i)
            var_y1 = 'var{}_y'.format(i + 1)
            var_z1 = 'var{}_z'.format(i + 1)

            sig_y = torch.unsqueeze(
                torch.tanh(getattr(self, name_y)(vars()[var_z0])), 2)
            K_y = torch.transpose(getattr(self, name_y).weight, 0, 1)
            vars()[var_y1] = torch.add(
                vars()[var_y0],
                1 * self.dh * torch.squeeze(torch.matmul(K_y, sig_y), 2))

            sig_z = torch.unsqueeze(
                torch.tanh(getattr(self, name_z)(vars()[var_y1])), 2)
            K_z = torch.transpose(getattr(self, name_z).weight, 0, 1)
            vars()[var_z1] = torch.add(
                vars()[var_z0],
                -1 * self.dh * torch.squeeze(torch.matmul(K_z, sig_z), 2))

        return torch.cat((vars()[var_y1], vars()[var_z1]), 1)

    def customized_loss(self, inputs, mapped_inputs, gradients):
        """
        Custom loss function.

        :param torch.Tensor inputs: DoubleTensor n_samples-by-n_params
            containing the points in the full input space.
        :param torch.Tensor mapped_inputs: DoubleTensor
            n_samples-by-n_params containing the mapped points in the
            full input space. They are the result of the forward application.
        :param torch.Tensor gradients: DoubleTensor n_samples-by-n_params
            containing the gradient samples wrt the input parameters.
        """
        # Define the weights and bias of the inverse network
        for i in range(self.n_layers):
            name_y = 'fc{}_y'.format(i + 1)
            name_z = 'fc{}_z'.format(i + 1)

            inv_name_y_weight = 'inv_fc{}_y_weight'.format(i + 1)
            inv_name_z_weight = 'inv_fc{}_z_weight'.format(i + 1)
            vars()[inv_name_y_weight] = getattr(self, name_y).weight
            vars()[inv_name_z_weight] = getattr(self, name_z).weight

            inv_name_y_bias = 'inv_fc{}_y_bias'.format(i + 1)
            inv_name_z_bias = 'inv_fc{}_z_bias'.format(i + 1)
            vars()[inv_name_y_bias] = getattr(self, name_y).bias
            vars()[inv_name_z_bias] = getattr(self, name_z).bias

        Jacob = torch.empty(inputs.size()[0], 2 * self.n_params,
                            2 * self.n_params)

        for j in range(2 * self.n_params):
            output_dy = torch.clone(mapped_inputs)
            output_dy[:, j] += 0.001

            bb = torch.split(output_dy, self.n_params, dim=1)
            var_y0 = 'var{}_y'.format(self.n_layers - 1)
            var_z0 = 'var{}_z'.format(self.n_layers - 1)
            vars()[var_y0] = bb[0]
            vars()[var_z0] = bb[1]

            for i in range(self.n_layers - 1, -1, -1):
                inv_name_y_weight = 'inv_fc{}_y_weight'.format(i + 1)
                inv_name_z_weight = 'inv_fc{}_z_weight'.format(i + 1)
                inv_name_y_bias = 'inv_fc{}_y_bias'.format(i + 1)
                inv_name_z_bias = 'inv_fc{}_z_bias'.format(i + 1)
                var_y0 = 'var{}_y'.format(i)
                var_z0 = 'var{}_z'.format(i)
                var_y1 = 'var{}_y'.format(i - 1)
                var_z1 = 'var{}_z'.format(i - 1)

                sig_z = torch.tanh(
                    torch.add(
                        torch.matmul(vars()[inv_name_z_weight],
                                     torch.unsqueeze(vars()[var_y0], 2)),
                        torch.unsqueeze(vars()[inv_name_z_bias], 1)))
                K_z = torch.transpose(vars()[inv_name_z_weight], 0, 1)
                vars()[var_z1] = torch.add(
                    vars()[var_z0],
                    1 * self.dh * torch.squeeze(torch.matmul(K_z, sig_z), 2))

                sig_y = torch.tanh(
                    torch.add(
                        torch.matmul(vars()[inv_name_y_weight],
                                     torch.unsqueeze(vars()[var_z1], 2)),
                        torch.unsqueeze(vars()[inv_name_y_bias], 1)))
                K_y = torch.transpose(vars()[inv_name_y_weight], 0, 1)
                vars()[var_y1] = torch.add(
                    vars()[var_y0],
                    -1 * self.dh * torch.squeeze(torch.matmul(K_y, sig_y), 2))

            dx = torch.cat((vars()[var_y1], vars()[var_z1]), 1)

            # # Test the invertibility
            # # Warning: it is computational intense
            # if torch.mean(torch.abs(torch.add(-1 * output_dy,
            #                                   self.forward(dx)))) > 1e-5:
            #     print('Something is wrong in Jacobian computation')
            #     print(torch.mean(
            #         torch.abs(torch.add(-1 * output_dy, self.forward(dx)))))

            for k in range(2 * self.n_params):
                Jacob[:, j, k] = torch.add(dx[:, k], -1 * inputs[:, k])

        JJ2 = torch.unsqueeze(
            torch.sqrt(torch.sum(torch.mul(Jacob, Jacob), 2)), 2)
        JJJ = torch.div(Jacob, JJ2.expand(-1, -1, 2 * self.n_params))
        ex_data = torch.unsqueeze(gradients, 2)
        loss_weights = torch.clone(torch.squeeze(torch.matmul(JJJ, ex_data),
                                                 2))
        # anisotropy weigths
        loss_weights[:, self.omega] = 0.0
        loss_anisotropy = torch.sqrt(
            torch.mean(torch.sum(torch.mul(loss_weights, loss_weights), 1)))

        J_det = torch.empty(inputs.shape[0])
        for k in range(inputs.shape[0]):
            eee = torch.svd(JJJ[k, :, :])[1]
            J_det[k] = torch.prod(eee)
        loss_det = torch.abs(torch.prod(J_det - 1.0))
        return loss_anisotropy + loss_det


class BackwardNet(nn.Module):
    """Backward Net class. It is part of the RevNet.

    :param int n_params: number of input parameters.
    :param int n_layers: number of layers of the RevNet.
    :param float dh: so-called time step of the RevNet.
    """
    def __init__(self, n_params, n_layers, dh):
        super().__init__()
        self.n_params = n_params // 2
        self.n_layers = n_layers
        self.dh = dh

        for i in range(self.n_layers):
            setattr(self, 'fc{}_y'.format(i + 1),
                    nn.Linear(self.n_params, 2 * self.n_params))
            setattr(self, 'fc{}_z'.format(i + 1),
                    nn.Linear(self.n_params, 2 * self.n_params))

    def forward(self, mapped_inputs):
        """
        Maps transformed inputs to original inputs.

        :param torch.Tensor mapped_inputs: DoubleTensor n_samples-by-n_params
            containing the nonlinear transformed inputs.
        :return inputs: DoubleTensor n_samples-by-n_params with
            the points in the original full input space.
        :rtype: torch.Tensor
        """
        y, z = torch.split(mapped_inputs, self.n_params, dim=1)

        for i in range(self.n_layers - 1, -1, -1):
            name_y = 'fc{}_y'.format(i + 1)
            name_z = 'fc{}_z'.format(i + 1)

            sig_z = torch.unsqueeze(torch.tanh(getattr(self, name_z)(y)), 2)
            K_z = torch.transpose(getattr(self, name_z).weight, 0, 1)
            z = torch.add(
                z, 1 * self.dh * torch.squeeze(torch.matmul(K_z, sig_z), 2))

            sig_y = torch.unsqueeze(torch.tanh(getattr(self, name_y)(z)), 2)
            K_y = torch.transpose(getattr(self, name_y).weight, 0, 1)
            y = torch.add(
                y, -1 * self.dh * torch.squeeze(torch.matmul(K_y, sig_y), 2))

        return torch.cat((y, z), 1)



"""Utility functions module.
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import linprog
import GPy

class Normalizer():
    """A class for normalizing and unnormalizing bounded inputs.

    :param numpy.ndarray lb: array n_params-by-1 that contains lower bounds
        on the simulation inputs.
    :param numpy.ndarray ub: array n_params-by-1 that contains upper bounds
        on the simulation inputs.
    """
    def __init__(self, lb, ub):
        self.lb = lb
        self.ub = ub

    def fit_transform(self, inputs):
        """Return corresponding points shifted and scaled to [-1, 1]^n_params.

        :param numpy.ndarray inputs: contains all input points to normalize.
            The shape is n_samples-by-n_params. The components of each row of
            `inputs` should be between `self.lb` and `self.ub`.
        :return: the normalized inputs. The components of each row should be
            between -1 and 1.
        :rtype: numpy.ndarray
        """
        inputs_norm = 2.0 * (inputs - self.lb) / (self.ub - self.lb) - 1.0
        return inputs_norm

    def inverse_transform(self, inputs):
        """Return corresponding points shifted and scaled to
        `[self.lb, self.ub]`.

        :param numpy.ndarray inputs: contains all input points to unnormalize.
            The shape is n_samples-by-n_params. The components of each row of
            `inputs` should be between -1 and 1.
        :return: the unnormalized inputs. The components of each row should be
            between `self.lb` and `self.ub`.
        :rtype: numpy.ndarray
        """
        inputs_unnorm = (self.ub - self.lb) * (inputs + 1.0) / 2.0 + self.lb
        return inputs_unnorm


def initialize_weights(matrix):
    """
    Inizialize uniform weights for simple Monte Carlo method or linear regression in
    local linear gradients.

    :param numpy.ndarray matrix: matrix which shape[0] value contains the
        dimension of the weights to be computed.
    :return: weights
    :rtype: numpy.ndarray
    """
    return np.ones((matrix.shape[0], 1)) / matrix.shape[0]


def linear_program_ineq(c, A, b):
    """Solves an equality constrained linear program with variable bounds.
    This method returns the minimizer of the following linear program.

    minimize  c^T x
    subject to  A x >= b

    :param numpy.ndarray c: coefficients vector of the linear objective
        function to be minimized.
    :param numpy.ndarray A: 2-D array which, when matrix-multiplied by x,
        gives the values of the lower-bound inequality constraints at x.
    :param numpy.ndarray b: 1-D array of values representing the lower-bound
        of each inequality constraint (row) in A.
    :return: the independent variable vector which minimizes the linear
        programming problem.
    :rtype: numpy.ndarray
    :raises: RuntimeError
    """
    c = c.reshape(-1, )
    b = b.reshape(-1, )

    # make unbounded bounds
    bounds = [(None, None) for i in range(c.shape[0])]
    res = linprog(c=c, A_ub=-A, b_ub=-b, bounds=bounds)
    if res.success:
        return res.x.reshape(-1, 1)
    raise RuntimeError('Scipy did not solve the LP. {}'.format(res.message))


def local_linear_gradients(inputs, outputs, weights=None, n_neighbors=None):
    """Estimate a collection of gradients from input/output pairs.

    Given a set of input/output pairs, choose subsets of neighboring points and
    build a local linear model for each subset. The gradients of these local
    linear models comprise estimates of sampled gradients.

    :param numpy.ndarray inputs:
        M-by-m matrix that contains the m-dimensional inputs
    :param numpy.ndarray outputs:
        M-by-1 matrix that contains scalar outputs
    :param numpy.ndarray weights:
        M-by-1 matrix that contains the weights for each observation (default
        None)
    :param int n_neighbors:
        how many nearest neighbors to use when constructing the local linear
        model. the default value is floor(1.7*m)
    :return: M-by-m matrix that contains estimated partial derivatives approximated
        by the local linear models; the corresponding new inputs
    :rtype: numpy.ndarray, numpy.ndarray

    :raises: ValueError, TypeError
    """
    n_samples, n_pars = inputs.shape

    if n_samples <= n_pars:
        raise ValueError('Not enough samples for local linear models.')

    if n_neighbors is None:
        n_neighbors = int(min(np.floor(1.7 * n_pars), n_samples))
    elif not isinstance(n_neighbors, int):
        raise TypeError(
            'n_neighbors ({}) must be an integer.'.format(n_neighbors))

    if n_neighbors <= n_pars or n_neighbors > n_samples:
        raise ValueError(
            'n_neighbors must be between the number of parameters '
            'and the number of samples. Unsatisfied: {} < {} < {}.'.format(
                n_pars, n_neighbors, n_samples))

    if weights is None:
        weights = initialize_weights(inputs)

    MM = min(int(np.ceil(10 * n_pars * np.log(n_pars))), n_samples - 1)

    # distinguish between scalar and vectorial outputs
    if len(outputs.shape) == 1 or outputs.shape[1] == 1:
        gradients = np.zeros((MM, n_pars))
    else:
        gradients = np.zeros((MM, outputs.shape[1], n_pars))

    # new inputs are defined since MM is different from n_samples
    new_inputs = np.zeros((MM, n_pars))

    for i in range(MM):
        ii = np.random.randint(n_samples)
        inputs_rand_row = inputs[ii, :]
        D2 = np.sum((inputs - inputs_rand_row)**2, axis=1)
        ind = np.argsort(D2)
        ind = ind[D2 != 0]
        A = np.hstack((np.ones(
            (n_neighbors, 1)), inputs[ind[:n_neighbors], :])) * np.sqrt(
                weights[ii])
        b = outputs[ind[:n_neighbors]] * np.sqrt(weights[ii])
        u = np.linalg.lstsq(A, b, rcond=None)[0]
        gradients[i] = u[1:].T
        new_inputs[i, :] = (1 / n_neighbors) * np.sum(
            inputs[ind[:n_neighbors], :], axis=0)
    return gradients, new_inputs


def sort_eigpairs(matrix):
    """Compute eigenpairs and sort.

    :param numpy.ndarray matrix: matrix whose eigenpairs you want.
    :return: vector of sorted eigenvalues; orthogonal matrix of corresponding
        eigenvectors.
    :rtype: numpy.ndarray, numpy.ndarray

    .. note::

        Eigenvectors are unique up to a sign. We make the choice to normalize
        the eigenvectors so that the first component of each eigenvector is
        positive. This normalization is very helpful for the bootstrapping.
    """
    evals, evects = np.linalg.eigh(matrix)
    evals = abs(evals)
    ind = np.argsort(evals)
    evals = evals[ind[::-1]]
    evects = evects[:, ind[::-1]]
    s = np.sign(evects[0, :])
    s[s == 0] = 1
    evects *= s
    return evals.reshape(-1, 1), evects


class CrossValidation():
    """
    Class to perform k-fold cross validation when tuning hyperparameters for the
    design of a response surface with ActiveSubspaces or KernelActiveSubspaces.
    Used in particular during the tuning of the parameters of the spectral
    distribution of the feature map, inside the object function average_rrmse.
    default score is the relative root mean square error (rrmse).

    :param numpy.ndarray inputs: n_samples-by-input_dim input matrix.
    :param numpy.ndarray outputs: n_sample-by-output_dim output matrix.
    :param numpy.ndarray gradients: n_samples-by-output_dim-by-input_dim
        gradients matrix.
    :param `Subspaces` subspace: ActiveSubspace or KernelActiveSubspace object,
        from which evaluate the response surface. The dimension of the response
        surface is specified in subspace.dim attribute.
    :param int folds: number of folds of the cross-validation procedure.
    :param dict kwargs: additional paramters organized in a dictionary to pass
        to subspace.fit method. For example 'weights' or 'metric'.

    :cvar `GPy.models.GPRegression` gp: Gaussian process of the response surface
        built with GPy library.
    """
    def __init__(self, inputs, outputs, gradients, subspace, folds=5, **kwargs):

        if any([v is None for v in [inputs, outputs, gradients, subspace]]):
            raise ValueError(
                'Any among inputs, outputs, gradients, subspace is None.'
            )

        self.inputs = inputs
        self.outputs = outputs
        self.gradients = gradients
        self.ss = subspace
        self.folds = folds
        self.gp = None
        self.kwargs = kwargs

    def run(self):
        """
        Run the k-fold cross validation procedure. In each fold a fit and an
        evaluation of the score are compute.

        :return: mean and standard deviation of the scores.
        :rtype: list of two numpy.ndarray.
        """
        mask = np.arange(self.inputs.shape[0])

        np.random.shuffle(mask)
        scores = np.zeros((self.folds))
        s_mask = np.array_split(mask, self.folds)

        for i in range(self.folds):
            v_mask = s_mask[i]
            validation = (self.inputs[v_mask, :], self.outputs[v_mask, :])

            t_mask = ~v_mask
            self.fit(self.inputs[t_mask, :], self.gradients[t_mask, :, :],
                     self.outputs[t_mask, :])

            scores[i] = self.scorer(validation[0], validation[1])

        return scores.mean(), scores.std()

    def fit(self, inputs, gradients, outputs):
        """
        Uses Gaussian process regression to build the response surface as a side
        effect. The dimension of the response surface is specified in the
        attribute self.ss.dim.

        :param numpy.ndarray inputs: n_samples-by-input_dim input matrix.
        :param numpy.ndarray outputs: n_sample-by-output_dim output matrix.
        :param numpy.ndarray gradients: n_samples-by-output_dim-by-input_dim
            gradients matrix.
        """
        self.ss.fit(inputs=inputs,
                    gradients=gradients,
                    outputs=outputs,
                    **self.kwargs)
        y = self.ss.transform(inputs)[0]

        # build response surface
        kern = GPy.kern.RBF(input_dim=y.shape[1], ARD=True)
        self.gp = GPy.models.GPRegression(y, np.atleast_2d(outputs), kern)
        self.gp.optimize_restarts(15, verbose=False)

    def predict(self, inputs):
        """
        Predict method of cross-validation.

        :param numpy.ndarray inputs: n_samples-by-input_dim input matrix.
        :return: n_samples-by-dim prediction of the surrogate response surface
            model at the inputs. The value dim corresponds to self.ss.dim.
        :rtype: numpy.ndarray
        """
        x_test = self.ss.transform(inputs)[0]
        y = self.gp.predict(np.atleast_2d(x_test))[0]
        return y

    def scorer(self, inputs, outputs):
        """
        Score function of cross-validation.

        :param numpy.ndarray inputs: n_samples-by-input_dim input matrix.
        :param numpy.ndarray outputs: n_sample-by-output_dim output matrix.
        :return: relative root mean square error between inputs and outputs.
        :rtype: np.float64
        """
        y = self.predict(inputs)
        return rrmse(y, outputs)


def rrmse(predictions, targets):
    """
    Evaluates the relative root mean square error. It can be vectorized for
    multidimensional predictions and targets.

    :param numpy.ndarray predictions: predictions input.
    :param numpy.ndarray targets: targets input.
    :return: relative root mean squared error
    :rtype: np.float64
    """
    n_samples = predictions.shape[0]
    if n_samples != targets.shape[0]:
        raise ValueError('Predictions and targets differ in number of samples.')

    t = np.atleast_2d(targets).reshape(n_samples, -1)
    p = np.atleast_2d(predictions).reshape(n_samples, -1)
    std_deviation = np.linalg.norm(t - np.mean(t, axis=0).reshape(1, -1))
    return np.linalg.norm(p - t) / std_deviation


def average_rrmse(hyperparams, best, csv, verbose=False, resample=5):
    """
    Objective function to be optimized during the tuning process of the method
    :func:`~athena.FeatureMap.tune_pr_matrix`. The optimal hyperparameters of the
    spectral distribution are searched for in a domain logarithmically scaled in
    base 10. For each call of :func:`~athena.utils.average_rrmse` by the
    optimizer, the same hyperparameter is tested in two nested procedures: in
    the external procedure the projection matrix is resampled a number of times
    specified by the resample parameter; in the internal procedure the relative
    root mean squared error (:func:`~athena.utils.rrmse`) is evaluated as the
    k-fold mean of a k-fold cross-validation procedure. The score of a single
    fold of this cross-validation procedure is the rrmse on the validation set
    of the predictions of the response surface built with a Subspace object on
    the training set.

    :param list hyperparameters: logarithm of the parameter of the spectral
        distribution passed to average_rrmse by the optimizer.
    :param 'CrossValidation' csv: CrossValidation object which contains the
        same Subspace object and the inputs, outputs, gradients datasets. The
    :param list best: list that records the best score and the best
        projection matrix. The initial values are 0.8 and a
        n_features-by-input_dim numpy.ndarray of zeros.
    :param int resample: number of times the projection matrix is resampled
        from the same spectral distribution with the same hyperparameter.
    :param bool verbose: True to print the score for each resample.
    :return: minumum of the scores evaluated for the same hyperparameter and
        a specified number of resamples of the projection matrix.
    :rtype: numpy.float64
    """
    if isinstance(csv, CrossValidation) is False:
        raise ValueError("The argument csv must be of type athena.utils.CrossValidation")

    if len(hyperparams.shape) > 1:
        hyperparams = np.squeeze(hyperparams)
    if len(hyperparams.shape) == 0:
        hyperparams = np.array([hyperparams])

    # compute the real hyperparameters
    hyperparams = 10**hyperparams

    # list of scores for the same hyperparameters but different samples
    # of the projection matrix
    score_records = []

    #set the hyperparameters
    csv.ss.feature_map.params = hyperparams

    if verbose is True:
        print("#" * 80)
    for _ in range(resample):
        # compute the score with cross validation for the sampled projection
        # matrix
        mean, std = csv.run()

        # save the best parameters
        if verbose is True:
            print("params {} mean {}, std {}".format(hyperparams, mean, std))
        score_records.append(mean)

        # skip resampling from the same hyperparam if the error is not below
        # the treshold 0.8
        if mean > 0.8:
            break
        if mean <= best[0]:
            best[0] = mean
            best[1] = csv.ss.feature_map.pr_matrix

        # set _pr_matrix to None so that csv.ss.feature_map.compute_fmap
        # and csv.ss.feature_map.compute_fmap_jac resample the projection matrix
        # for the same hyperparams
        csv.ss.feature_map._pr_matrix = None
    return min(score_records)



"""
Athena init
"""
__all__ = [
    'active', 'feature_map', 'kas', 'projection_factory', 'nll', 'subspaces',
    'utils'
]

from .active import ActiveSubspaces
from .feature_map import (FeatureMap, rff_map, rff_jac)
from .kas import KernelActiveSubspaces
from .projection_factory import ProjectionFactory
from .nll import NonlinearLevelSet, ForwardNet, BackwardNet
from .subspaces import Subspaces
from .utils import (Normalizer, initialize_weights, linear_program_ineq,
                    local_linear_gradients, sort_eigpairs, CrossValidation,
                    rrmse, average_rrmse)



"""
Package metadata
"""
__all__ = [
    '__project__',
    '__title__',
    '__author__',
    '__copyright__',
    '__license__',
    '__version__',
    '__mail__',
    '__maintainer__',
    '__status__',
    '__packagename__']

__project__ = 'ATHENA'
__title__ = "athena"
__author__ = "Marco Tezzele, Francesco Romor"
__copyright__ = "Copyright 2019-2021, Athena contributors"
__license__ = "MIT"
__version__ = "0.1.1"
__mail__ = 'marcotez@gmail.com, francesco.romor@gmail.com'
__maintainer__ = __author__
__status__ = "Stable"
__packagename__ = "athena-mathlab"



from setuptools import setup

meta = {}
with open("athena/meta.py") as fp:
    exec(fp.read(), meta)

# Package meta-data.
IMPORTNAME = meta['__title__']
PIPNAME = meta['__packagename__']
DESCRIPTION = 'Advanced Techniques for High dimensional parameter spaces to ' \
              'Enhance Numerical Analysis'
URL = 'https://github.com/mathLab/ATHENA'
MAIL = meta['__mail__']
AUTHOR = meta['__author__']
VERSION = meta['__version__']
KEYWORDS = 'parameter-space-reduction active-subspaces ' \
           'kernel-active-subspaces model-reduction sensitivity-analysis ' \
           'nonlinear-level-set-learning'

REQUIRED = ['numpy', 'scipy', 'matplotlib', 'torch', 'GPy', 'GPyOpt', 'scikit-learn']

EXTRAS = {
    'docs': ['Sphinx>=1.4', 'sphinx_rtd_theme'],
    'formatting': ['yapf'],
    'tutorials': ['pyro', 'pyhmc'],
}

LDESCRIPTION = ('ATHENA is a Python package for reduction in parameter spaces.')

setup(name=PIPNAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LDESCRIPTION,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Mathematics'
      ],
      keywords=KEYWORDS,
      url=URL,
      author=AUTHOR,
      author_email=MAIL,
      license='MIT',
      packages=[IMPORTNAME],
      install_requires=REQUIRED,
      extras_require=EXTRAS,
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      zip_safe=False)



"""
Module for the factory class for projection matrices
"""
import numpy as np


class classproperty():
    """
    Custom decorator.
    """
    def __init__(self, f):
        self.f = f
        self.__doc__ = f.__doc__
        self.__name__ = f.__name__

    def __get__(self, obj, owner):
        return self.f(owner)


class ProjectionFactory():
    """
    Factory class that spawns projection matrices.

    :Example:

        >>> from athena import ProjectionFactory
        >>> import numpy as np
        >>> input_dim = 2
        >>> n_features = 40
        >>> params = [1.1, 2.9]
        >>> for pname in ProjectionFactory.projections:
        >>>     y = ProjectionFactory(pname)(input_dim, n_features, params)
    """
    @staticmethod
    def beta(input_dim, n_features, params):
        """
        Beta distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the two parameters are the alpha and beta
            shape parameters respectively.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return np.random.beta(params[0], params[1], (n_features, input_dim))

    @staticmethod
    def cauchy(input_dim, n_features, params):
        """
        Cauchy distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the single parameter is a scale factor.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return (1 / params[0]) * np.random.standard_cauchy(
            (n_features, input_dim))

    @staticmethod
    def dirichlet(input_dim, n_features, params):
        """
        Dirichlet distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the single parameter is a scale to the input_dim
            dimensional shape parameter.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return np.random.dirichlet(params[0] * np.ones(input_dim), n_features)

    @staticmethod
    def laplace(input_dim, n_features, params):
        """
        Laplace distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the single parameter is the scale of the
            distribution, the mean is set to 0.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return np.random.laplace(0, params[0], (n_features, input_dim))

    @staticmethod
    def multivariate_normal(input_dim, n_features, params):
        """
        Multivariate normal distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the input_dim dimensioanl parameters are the
        diagonal of the covariance matrix of the distribution. The mean is set
        to the 0 vector.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return np.random.multivariate_normal(np.zeros(input_dim),
                                             np.diag(params), n_features)

    @staticmethod
    def normal(input_dim, n_features, params):
        """
        Normal distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the single parameter is the variance of the
            distribution. The mean is set to 0.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return np.random.normal(0, params[0], (n_features, input_dim))

    @staticmethod
    def uniform(input_dim, n_features, params):
        """
        Uniform distribution

        :param int input_dim: dimension of the inputs.
        :param int n_features: dimension of the RKHS.
        :param list params: the two parameters are the extremals of the domain.
        :return: n_features-by-input_dim projection matrix.
        :rtype: numpy.ndarray.
        """
        return np.random.uniform(params[0], params[1], (n_features, input_dim))

    ###########################################################################
    ##                                                                       ##
    ## PROJECTION FUNCTION dictionary                                        ##
    ##                                                                       ##
    ## New implementations must be added here.                               ##
    ##                                                                       ##
    ###########################################################################
    __projections = {
        'beta': beta.__func__,
        'cauchy': cauchy.__func__,
        'dirichlet': dirichlet.__func__,
        'laplace': laplace.__func__,
        'multivariate_normal': multivariate_normal.__func__,
        'normal': normal.__func__,
        'uniform': uniform.__func__
    }

    def __new__(self, fname):
        # to make the str callable we have to use a dictionary with all the
        # implemented projection matrices
        if fname in self.projections:
            return self.__projections[fname]
        raise NameError(
            """The name of the projection matrix is not correct or not
            implemented. Check the documentation for all the available
            possibilities.""")

    @classproperty
    def projections(self):
        """
        The available projection matrices.

        :return: the list of all the available projection matrices.
        :rtype: list
        """
        return list(self.__projections.keys())



"""
Module for Kernel-based Active Subspaces.

:References:

    - Francesco Romor, Marco Tezzele, Andrea Lario, Gianluigi Rozza.
      Kernel-based Active Subspaces with application to CFD problems using
      Discontinuous Galerkin Method. 2020.
      arxiv: https://arxiv.org/abs/2008.12083

"""
import numpy as np
from .subspaces import Subspaces
from .utils import (initialize_weights, local_linear_gradients)
from .feature_map import FeatureMap


class KernelActiveSubspaces(Subspaces):
    """
    Compute the kernel-based active subspaces given the inputs and the
    gradients of the model function wrt the input parameters, or given the
    input/outputs couples. Only two methods are available: 'exact' and 'local'.

    :param `FeatureMap` feature_map: `athena.feature_map.FeatureMap` object. See
        :class:`FeatureMap` class in :py:mod:`feature_map` module.
    :param int n_features: dimension of the feature space.
    :param str method: method to compute the AS. Possible choices are
        'exact' when the gradients are provided, or 'local' to use local linear
        models. This approach is related to the sufficient dimension reduction
        method known sometimes as the outer product of gradient method. See the
        2001 paper 'Structure adaptive approach for dimension reduction' from
        Hristache, et al.
    :param int n_boot: number of bootstrap samples. Default is None.

    :cvar numpy.ndarray inputs: input parameters oriented as rows.
    :cvar numpy.ndarray outputs: corresponding outputs oriented as rows.
    :cvar numpy.ndarray gradients: n_samples-by-n_params matrix containing the
        gradient samples oriented as rows.
    :cvar numpy.ndarray weights: n_samples-by-1 weight vector, corresponds
        to numerical quadrature rule used to estimate matrix whose eigenspaces
        define the active subspace.
    :cvar int n_features: dimension of the feature space.
    :cvar numpy.ndarray features: n_samples-by-n_features matrix containing
        the projections of the inputs to the n_features-dimensional feature
        space.
    :cvar numpy.ndarray pseudo_gradients:
    :cvar str method: method to compute the AS. Possible choices are
        'exact' when the gradients are provided, or 'local' to use local linear
        models. This approach is related to the sufficient dimension reduction
        method known sometimes as the outer product of gradient method. See the
        2001 paper 'Structure adaptive approach for dimension reduction' from
        Hristache, et al.
    :cvar int n_boot: number of bootstrap samples.
    :cvar numpy.ndarray metric: metric matrix for vectorial active
        subspaces.
    """
    def __init__(self,
                 dim,
                 feature_map=None,
                 n_features=None,
                 method='exact',
                 n_boot=None):
        super().__init__(dim, method, n_boot)
        self.feature_map = feature_map
        self.n_features = n_features
        self.features = None
        self.pseudo_gradients = None

    def fit(self,
            inputs=None,
            outputs=None,
            gradients=None,
            weights=None,
            metric=None):
        """
        Compute the kernel based active subspaces given the inputs and the
        gradients of the model function wrt the input parameters, or given the
        input/outputs couples. Only two methods are available: 'exact' and
        'local'.

        :param numpy.ndarray inputs: array n_samples-by-n_params containing
            the points in the original parameter space.
        :param numpy.ndarray outputs: array n_samples-by-1 containing
            the values of the model function.
        :param numpy.ndarray gradients: array n_samples-by-n_params containing
            the gradient samples oriented as rows.
        :param numpy.ndarray weights: n_samples-by-1 weight vector, corresponds
            to numerical quadrature rule used to estimate matrix whose
            eigenspaces define the active subspace.
        :param numpy.ndarray metric: metric matrix output_dim-by-output-dim for
            vectorial active subspaces.
        :raises: TypeError, ValueError
        """
        if self.method == 'exact':
            if gradients is None or inputs is None:
                raise TypeError('gradients or inputs argument is None.')

        # estimate active subspace with local linear models.
        elif self.method == 'local':
            if inputs is None or outputs is None:
                raise TypeError('inputs or outputs argument is None.')
            gradients, inputs = local_linear_gradients(inputs=inputs,
                                                       outputs=outputs,
                                                       weights=weights)

        if len(gradients.shape) == 2:
            gradients = gradients.reshape(gradients.shape[0], 1,
                                          gradients.shape[1])

        if weights is None or self.method == 'local':
            # use the new gradients to compute the weights, otherwise dimension
            # mismatch accours.
            weights = initialize_weights(gradients)

        if self.n_features is None:
            self.n_features = inputs.shape[1]

        if self.feature_map is None:
            # default spectral measure is Gaussian
            self.feature_map = FeatureMap(distr='multivariate_normal',
                                          bias=np.ones((1, self.n_features)),
                                          input_dim=inputs.shape[1],
                                          n_features=self.n_features,
                                          params=np.ones(inputs.shape[1]),
                                          sigma_f=1)

        if len(gradients.shape) == 3 and metric is None:
            metric = np.diag(np.ones(gradients.shape[1]))

        self.pseudo_gradients, self.features = self._reparametrize(
            inputs, gradients)

        self.evals, self.evects = self._build_decompose_cov_matrix(
            self.pseudo_gradients, weights, metric)

        if self.n_boot:
            if self.n_boot <= 50:
                self._compute_bootstrap_ranges(gradients=self.pseudo_gradients,
                                               weights=weights,
                                               metric=metric)
            else:
                raise ValueError(
                    'n_boot is too high for the bootstrap method applied to kas'
                )

        self._partition()

    def transform(self, inputs):
        """
        Map full variables to active and inactive variables.
        Points in the original input space are mapped to the active and
        inactive non-linear subspace.

        :param numpy.ndarray inputs: array n_samples-by-n_params containing
            the points in the original parameter space.
        :return: array n_samples-by-active_dim containing the mapped active
            variables;
            array n_samples-by-inactive_dim containing the mapped inactive
            variables.
        :rtype: numpy.ndarray, numpy.ndarray
        :raises: TypeError
        """
        features = self.feature_map.compute_fmap(inputs)

        if self.W1 is None:
            raise TypeError(
                'the kernel-based active subspace has not been evaluated.')

        # allow evaluation of active variables only
        if self.W2 is None:
            active = np.dot(features, self.W1)
            inactive = None
        else:
            active = np.dot(features, self.W1)
            inactive = np.dot(features, self.W2)

        return active, inactive

    def inverse_transform(self, reduced_inputs, n_points):
        pass

    def _reparametrize(self, inputs, gradients):
        """
        Computes the pseudo-gradients solving an overdetermined linear system.

        :param numpy.ndarray inputs: array n_samples-by-n_params containing
            the points in the original parameter space.
        :param numpy.ndarray gradients: array n_samples-by-n_params containing
            the gradient samples oriented as rows.
        :return: array n_samples-by-output_dim-by-n_params matrix containing
            the pseudo gradients corresponding to each sample;
            array n_samples-by-n_features containing the image of the inputs
            in the feature space.
        :rtype: numpy.ndarray, numpy.ndarray
        """
        n_samples = inputs.shape[0]

        # Initialize Jacobian for each input
        jacobian = self.feature_map.compute_fmap_jac(inputs)

        # Compute pseudo gradients
        pseudo_gradients = np.array([
            np.linalg.lstsq(jacobian[i, :, :].T,
                            gradients[i, :, :].T,
                            rcond=None)[0].T for i in range(n_samples)
        ])

        # Compute features
        features = self.feature_map.compute_fmap(inputs)

        return pseudo_gradients, features
